from jesth.errors import Error
from jesth.converter import ValueConverter
from collections import OrderedDict


def validate(data, schema, value_converter=None):
    value_converter = value_converter if value_converter else ValueConverter()
    validator = Validator(schema, value_converter=value_converter)
    return validator.validate(data)


class Spec:
    def __init__(self, datatype, checker=None):
        self._datatype = datatype
        self._checker = checker

    @property
    def datatype(self):
        return self._datatype

    @property
    def checker(self):
        return self._checker


class Validator:
    def __init__(self, schema, value_converter=None):
        self._schema = schema
        self._value_converter = value_converter if value_converter else ValueConverter()

    @property
    def schema(self):
        return self._schema

    @property
    def value_converter(self):
        return  self._value_converter

    def validate(self, target):
        return self._validate(target, self._schema)

    def _validate(self, target, schema):
        if target is None:
            return True
        if isinstance(schema, str):
            return self._validate_datatype(target, schema)
        if isinstance(schema, Spec):
            return self._ensure_spec(target, schema)
        if isinstance(schema, list) or isinstance(schema, tuple):
            return self._validate_list(target, schema)
        if isinstance(schema, dict) or isinstance(schema, OrderedDict):
            return self._validate_dict(target, schema)
        msg = "Invalid schema"
        raise Error(msg)

    def _validate_list(self, target, schema):
        """Schema SHOULD be a list"""
        if type(target) not in self._value_converter.list_types:
            return False
        if not schema:
            return True
        for item in target:
            matched = False
            for expected_type in schema:
                if self._validate(item, expected_type):
                    matched = True
                    break
            if not matched:
                return False
        return True

    def _validate_dict(self, target, schema):
        if type(target) not in self._value_converter.dict_types:
            return False
        if not schema:
            return True
        for key, val in target.items():
            try:
                r = self._validate(val, schema[key])
            except KeyError as e:
                return False
            else:
                if not r:
                    return False
        return True

    def _validate_datatype(self, target, datatype):
        valid_types_map = {"dict": self._value_converter.dict_types,
                           "list": self._value_converter.list_types,
                           "bin": self._value_converter.bin_types,
                           "bool": self._value_converter.bool_types,
                           "complex": self._value_converter.complex_types,
                           "date": self._value_converter.date_types,
                           "datetime": self._value_converter.datetime_types,
                           "float": self._value_converter.float_types,
                           "int": self._value_converter.integer_types,
                           "str": (tuple(self._value_converter.string_types)
                                   + tuple(self._value_converter.raw_types)
                                   + tuple(self._value_converter.text_types)),
                           "time": self._value_converter.time_types}
        try:
            valid_types = valid_types_map[datatype]
        except KeyError as e:
            msg = "Unknown data type '{}'".format(datatype)
            raise Error(msg)
        for valid_type in valid_types:
            if type(target) == valid_type:
                return True
        return False

    def _ensure_spec(self, target, spec):
        if not self._validate_datatype(target, spec.datatype):
            return False
        if not spec.checker:
            return True
        try:
            return bool(spec.checker(target))
        except Exception as e:
            msg = "Error while running a checker"
            raise Error(msg)


import zlib
from jesth.dataformat.text import encode as jesth_encode, decode as jesth_decode


def serialize(data, value_converter=None):
    data = jesth_encode(data).encode("utf-8")
    return zlib.compress(data, 6)


def deserialize(raw, value_converter=None):
    data = zlib.decompress(raw).decode("utf-8")
    return jesth_decode(data)



import unittest
from jesth import box


class TestBoxesClasses(unittest.TestCase):

    def test_raw_string(self):
        r = box.RawString("hello world")
        self.assertIsInstance(r, str)

    def test_text(self):
        r = box.Text("hello world")
        self.assertIsInstance(r, str)

    def test_raw_text(self):
        r = box.RawText("hello world")
        self.assertIsInstance(r, str)

    def test_hex_int(self):
        r = box.HexInt(420)
        self.assertIsInstance(r, int)

    def test_oct_int(self):
        r = box.OctInt(42)
        self.assertIsInstance(r, int)

    def test_bin_int(self):
        r = box.BinInt(4)
        self.assertIsInstance(r, int)

    def test_comment_id(self):
        r1 = box.CommentID()
        r2 = box.CommentID()
        self.assertIsInstance(r1, int)
        self.assertIsInstance(r2, int)
        self.assertNotEqual(r1, r2)

    def test_comment(self):
        r = box.Comment("# this is a comment")
        self.assertIsInstance(r, str)

    def test_whitespace_id(self):
        r1 = box.WhitespaceID()
        r2 = box.WhitespaceID()
        self.assertIsInstance(r1, int)
        self.assertIsInstance(r2, int)
        self.assertNotEqual(r1, r2)

    def test_whitespace(self):
        r = box.Whitespace()
        self.assertIsInstance(r, str)


if __name__ == '__main__':
    unittest.main()



from jesth.parser import read
from jesth import misc
from jesth.converter import flatten_dict, create_dict


def load(source, value_converter=None):
    """
    Load data from a file or file object
    """
    doc = read(source, value_converter)
    if not doc.sections:
        return None
    section = doc.sections[0]
    if section.header != "":
        return None
    body = section.to_dict()
    if not body:
        return None
    return body.get("data")


def dump(data, destination, value_converter=None):
    body = encode(data, value_converter)
    misc.write_to_file(body, destination)


def encode(data, value_converter=None):
    body = {"data": data}
    body = flatten_dict(body, value_converter=value_converter)
    return "\n".join(body)


def decode(text, value_converter=None):
    data = create_dict(text, value_converter=value_converter)
    return data.get("data")



"""Definition of the Section class"""
from jesth import converter, errors
from jesth.converter import ValueConverter


class Section:
    """A section is the main unit of a jesthFile. It's composed of a header and a body"""
    def __init__(self, header, body=None, *, value_converter=None):
        """
        Init a section

        [parameters]
        - header: the header (string) of the section
        - body: the body of the section is a string, a list of strings or a dict.
        Beware, if the body is a dict, instantiating this class may raise exceptions.
        For example, if the body contains circular references, an exception will be raised.
        Use the method ".update" if you want to submit a dict as body
        - value_converter: instance of ValueConverter to customize value encoding
        """
        self._header = header
        self._body = list()
        self._value_converter = value_converter if value_converter else ValueConverter()
        # update body
        self.update(body)

    @property
    def header(self):
        return self._header

    @property
    def body(self):
        """The body in its base form, i.e., a list of strings !"""
        return self._body

    @property
    def value_converter(self):
        return self._value_converter

    def to_dict(self, default=None, fallback=None, *, strict=True):
        """
        Try to convert the body of this section to a dict.
        Return a fallback value if failed to convert

        [parameters]
        - default: Value to return if the body has been successfully
          converted into a dict but empty. Note that `default` will be
          updated to contain an empty dict (from ValueConverter.dict_type)
          if you leave it set to None.
        - fallback: Value to return if the attempt to convert the body into a dict failed
        - strict: set True if you don't want to preserve comments and whitespaces

        [return]
        If everything is ok, a dict will be returned.
        Else the value of `default` or `fallback` will be returned.
        """
        default = self._value_converter.dict_type() if default is None else default
        try:
            return self.create_dict(default=default, strict=strict)
        except Exception as e:
            return fallback

    def to_text(self):
        """
        Return the body (list of lines) as a text string.

        [return]
        Return a string built with: `"\n".join(self._body)`
        """
        return "\n".join(self._body)

    def create_dict(self, default=None, *, strict=True):
        """
        Try to build a dict object out of the body of this section.
        Raise an exception if an errors occurs !
        Return `default` if the body has been created but is empty

        [parameters]
        - default: Value to return if the body has been successfully
          converted into an empty dict.
          Note that `default` will be updated to contain
          an empty dict (from ValueConverter.dict_type)
          if you leave it set to None.
        - strict: set True if you don't want to preserve comments and whitespaces

        [return]
        If everything is ok, a dict will be returned or the value of `default`
        Exceptions will be raised whenever a problem will arise !
        """
        default = self._value_converter.dict_type() if default is None else default
        data = converter.create_dict(self._body,
                                     value_converter=self._value_converter,
                                     strict=strict)
        return data if data else default

    def update(self, body):
        """Update the entire body. The new body may be string, list of strings or a dict.
        Beware, if the body is a dict, instantiating this class may raise exceptions.
        For example, if the body contains circular references, an exception will be raised."""
        self._body = ensure_body(body,
                                 value_converter=self._value_converter)


def ensure_body(body, value_converter=None):
    """Convert the body in the canonical base form: a list of string."""
    body = body if body else list()
    # string
    if isinstance(body, str):
        lines = body.split("\n")
        return lines
    # dict
    if isinstance(body, dict):
        return converter.flatten_dict(body, value_converter=value_converter)
    # list
    if isinstance(body, list):
        body = [line.replace("\n", "\\n") for line in body]
        return body
    raise errors.ConversionError



import math
import struct
from jesth.converter import ValueConverter
from jesth.tagbyte import TagByte
from jesth import errors


VERSION = 1

"""
OPEN_LIST:      \x01            
CLOSE_LIST:     \x02           
INTEGER:        \x03 size data  
STRING:         \x04 size data


"""


def serialize(data):
    serializer = Serializer(data)
    return serializer.run()


class Serializer:
    def __init__(self, data, value_converter=None):
        self._data = data
        self._value_converter = value_converter if value_converter else ValueConverter()
        self._buffer = bytearray()
        self._setup()

    def run(self):
        # insert version here
        self._run(self._data)
        return bytes(self._buffer)

    def _setup(self):
        insert_version(self._buffer, VERSION)

    def _run(self, data):
        data_type = type(data)
        # integer
        if data_type in self._value_converter.integer_types:
            self._encode_integer(data)
        # list
        elif data_type in self._value_converter.list_types:
            self._encode_list(data)
        else:
            msg = "Unknown datatype {}".format(data_type)
            raise errors.Error(msg)

    def _encode_integer(self, data):
        cache = encode_integer(data)
        self._buffer.extend(cache)

    def _encode_list(self, data):
        # encode list head
        cache = encode_list_head()
        self._buffer.extend(cache)
        # iterate list
        for item in data:
            self._run(item)
        # encode list tail
        cache = encode_list_tail()
        self._buffer.extend(cache)



def encode_integer(data):
    tag = TagByte.INTEGER.value
    spec = "<cH"
    return struct.pack(spec, tag, data)


def encode_list_head():
    tag = TagByte.LIST_HEAD.value
    spec = "<c"
    return struct.pack(spec, tag)


def encode_list_tail():
    tag = TagByte.LIST_TAIL.value
    spec = "<c"
    return struct.pack(spec, tag)


def insert_version(buffer, n):
    spec = "<B"
    b = struct.pack(spec, n)
    buffer.extend(b)


def write_integer(buffer):
    spec = "<"
    buffer.extend(struct.pack(spec))



def calc_integer_size(x):
    """Return the number of bytes needed for a given integer"""
    sign_bit = 0 if x >= 0 else 1
    n_bits = sign_bit + x.bit_length()
    return math.ceil(n_bits / 8.0)



"""Document class for creating model for Jesth data or to interacting with a Jesthfile"""
import pathlib
from jesth import misc, validator
from jesth.section import Section
from jesth.renderer import write, render
from jesth.converter import ValueConverter
from jesth.errors import Error


class Document:
    """Create a model for Jesth data or for interacting with a JesthFile"""
    def __init__(self, path=None, sections=None, *, value_converter=None):
        """
        Init the document

        [parameters]
        - path: if the document is linked to a JesthFile,`path` contains the path to this file.
            Path might also be a pathlib.Path instance.
        - sections: list of sections (or tuple)
        - value_converter: an instance of jesth.converter.ValueConverter
        """
        self._path = str(path.resolve()) if isinstance(path, pathlib.Path) else path
        sections = sections if sections else list()
        self._sections = list(sections) if not isinstance(sections, list) else sections
        self._schema = None
        self._value_converter = value_converter if value_converter else ValueConverter()
        self._is_new = False
        self._model = None
        self._setup()

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, val):
        self._path = val

    @property
    def sections(self):
        return self._sections.copy()

    @sections.setter
    def sections(self, val):
        self._sections = val
        self._create_model()

    @property
    def schema(self):
        return self._schema

    @property
    def value_converter(self):
        return self._value_converter

    @property
    def headers(self):
        if not self._sections:
            return None
        return misc.get_headers(self._sections)

    def render(self, *headers, spacing=1):
        """
        Render the entire document or a specific set of sections, i.e.,
        returns a Jesth string that may be stored in a file.

        [parameters]
        - *headers: Headers of sections to render. Omitting this will render the entire section
        - spacing: number of empty lines between two adjacent sections. Defaults to 1 empty line.

        [return]
        Returns a string that contains sections (each made of square-brackets delimited header
        and associated body)
        """
        return render(self, *headers,
                      spacing=spacing)

    def save(self):
        """
        Save the recent modifications if this document is linked to a JesthFile,
         i.e., the path parameter has been set.

        [return]
        Returns True or False
         """
        if not self._path:
            return False
        write(self, destination=self._path)
        return True

    def save_to(self, path):
        """
        Save the document in a specific filename

        [parameters]
        - path: path to filename. Path may be a pathlib.Path instance
        """
        if isinstance(path, pathlib.Path):
            path = str(path.resolve())
        write(self, destination=path)

    def append(self, header, body=None):
        """
        Create a new section then append it to the end of this document

        [parameters]
        - header: the header (string) of the new section
        - body: the body of this section as a string or a list of strings
        """
        n = len(self._sections)
        self._insert(n, header, body)

    def insert(self, index, header, body=None):
        """
        Create a new section then insert it in the document at a specific index

        [parameters]
        - index: integer index (absolute index). Accepts -1 to mimic the "append" method
        - header: the header (string) of the new section
        - body: the body of this section as a string or a list of strings
        """
        n = len(self._sections)
        index = misc.correct_index(index, n,
                                   ignore_upper_bound=True) if n else 0
        self._insert(index, header, body)

    def set(self, index, header, body=None):
        """
        Set a section at a specific index of the document

        [parameters]
        - index: the index integer
        - header: the header string of the section
        - body: the body of the section, may be a list of string, a dictionary,
          or a text string
        """
        n = len(self._sections)
        index = misc.correct_index(index, n) if n else 0
        # find section at same index then remove it
        cache = self._find_section_to_remove(index)
        if cache:
            h, i = cache
            self.remove(header=h, sub_index=i)
        # insert new section
        self._insert(index, header, body)

    def get(self, header, sub_index=0):
        """
        Get X named section object located at Y index relatively to others sections with same header

        [parameters]
        - header: the header (string) of the section
        - index: integer index, relatively to the section
        family (sections sharing this same header).
        Defaults to 0, thus will be returned, the first section with this header relatively
        to this header family.

        [return]
        Returns a section or None
        """
        try:
            n = len(self._model[header])
        except KeyError:
            return None
        sub_index = misc.correct_index(sub_index, n) if n else 0
        section = None
        try:
            section = self._model[header][sub_index]
        except IndexError:
            pass
        return section

    def get_all(self, header):
        """
        Get all sections sharing same header

        [parameters]
        - header: the header (string) of the sections

        [return]
        Returns a list of sections whose headers match with the `header` parameter.
        """
        return self._model.get(header, list()).copy()

    def count(self, header):
        """
        Count the number of sections whose headers match with the
        `header` parameter

        [parameters]
        - header: the header (string) of the sections

        [return]
        Returns the integer number of sections matching with the `header` parameter
        """
        if header not in self._model:
            return 0
        return len(self._model[header])

    def bind_schema(self):  # TODO
        pass

    def unbind_schema(self):  # TODO
        pass

    def validate(self, *headers):  # TODO
        """
        Validate this document
        [parameters]
        - *headers: headers to validate. If you ignore this parameter, the document will
        be checked against the schema.
        [return]
        Return true if the document is valid. Raise an exception if the schema is missing
        """
        if self.schema is None:
            msg = "Missing schema"
            raise Error(msg)
        if not isinstance(self.schema, dict):
            msg = "The schema must be a dictionary whose keys represent the section headers"
            raise Error(msg)
        for key, val in self.schema.items():
            validator.validate()



        #validator.validate(self.)

    def remove(self, header, sub_index=-1):
        """
        Remove a section from this document

        [parameters]
        - header: the header of the section to remove
        - index: the index (integer) of the section relatively to
        its family (sections sharing same header).
        Defaults to -1, thus the last section of the given header family will be removed
        from the document
        """
        try:
            n = len(self._model[header])
        except KeyError:
            return
        sub_index = misc.correct_index(sub_index, n)
        if n == 0:
            del self._model[header]
        else:
            try:
                del self._model[header][sub_index]
            except IndexError:
                pass
        i1 = i2 = 0
        for section in self._sections:
            if section.header == header:
                if i2 == sub_index:
                    del self._sections[i1]
                    break
                i2 += 1
            i1 += 1

    def remove_all(self, header):
        """
        Remove all sections with this specific header

        [parameters]
        - header: the header of the section to remove
        """
        if header not in self._model:
            return
        del self._model[header]
        i = len(self._sections) - 1
        for section in reversed(self._sections):
            if section.header == header:
                del self._sections[i]
            i -= 1

    def _setup(self):
        self._create_model()

    def _create_model(self):
        self._model = dict()
        for section in self._sections:
            header = section.header
            if header not in self._model:
                self._model[header] = list()
            self._model[header].append(section)

    def _insert(self, index, header, body):
        section = Section(header, body=body,
                          value_converter=self._value_converter)
        self._sections.insert(index, section)
        self._update_model(header, index, section)

    def _update_model(self, header, index, section):
        if header not in self._model:
            self._model[header] = list()
        self._model[header].insert(index, section)

    def _find_section_to_remove(self, index):
        # find section at same index (the section to remove)
        header = None
        for i, section in enumerate(self._sections):
            if i == index:
                header = section.header
        if header is None:
            return
        i1 = i2 = 0
        for section in self._sections:
            if i1 == index:
                return header, i2
            if section.header == header:
                i2 += 1
            i1 += 1



import unittest
import tempfile
import pathlib
from decimal import Decimal
from jesth import Parser, parse, read
from jesth.converter import ValueConverter


TEXT = """\
Anonymous
section
here.

[section2]
hello
\\[world
"""


TEXT_WITH_END = """\
[section1]
first
section

[[END]]

[section2]
second
section
"""


class TestReadFunc(unittest.TestCase):

    def setUp(self):
        file = tempfile.NamedTemporaryFile(delete=False)
        file.write(TEXT.encode("utf-8"))
        file.close()
        self._path_str = file.name
        self._path_obj = pathlib.Path(self._path_str)

    def tearDown(self):
        self._path_obj.unlink()

    def test_read_from_file(self):
        with open(self._path_str, mode="rb") as file:
            document = read(file)
        test_document(document, self)

    def test_read_from_path_str(self):
        document = read(self._path_str)
        test_document(document, self)

    def test_read_from_path_obj(self):
        document = read(self._path_obj)
        test_document(document, self)

    def test_read_with_custom_value_converter(self):
        custom_float_decoder = Decimal
        custom_value_converter = ValueConverter(float_decoder=custom_float_decoder)
        document = read(self._path_obj, value_converter=custom_value_converter)
        section1 = document.sections[0]
        section2 = document.sections[1]
        expected = custom_float_decoder
        self.assertEqual(expected, section1.value_converter.float_decoder)
        self.assertEqual(expected, section2.value_converter.float_decoder)


class TestParseFunc(unittest.TestCase):

    def test_with_default_args(self):
        document = parse(TEXT)
        test_document(document, self)

    def test_with_custom_value_converter(self):
        custom_float_decoder = Decimal
        custom_value_converter = ValueConverter(float_decoder=custom_float_decoder)
        document = parse(TEXT, value_converter=custom_value_converter)
        section1 = document.sections[0]
        section2 = document.sections[1]
        expected = custom_float_decoder
        self.assertEqual(expected, section1.value_converter.float_decoder)
        self.assertEqual(expected, section2.value_converter.float_decoder)


class TestParserClass(unittest.TestCase):

    def test(self):
        parser = Parser()
        self.assertTrue(parser.feedable)
        for line in TEXT_WITH_END.split("\n"):
            parser.feed(line)
            if line.lower() == "[[end]]":
                self.assertFalse(parser.feedable)
        self.assertEqual(1, len(parser.sections))


def test_document(document, test_case):
    test_case.assertEqual(2, len(document.sections))
    section1, section2 = document.sections
    # section1
    expected_header = ""
    expected_body = ["Anonymous", "section", "here."]
    test_case.assertEqual(expected_header, section1.header)
    test_case.assertEqual(expected_body, section1.body)
    # section2
    expected_header = "section2"
    expected_body = ["hello", "[world"]
    test_case.assertEqual(expected_header, section2.header)
    test_case.assertEqual(expected_body, section2.body)


if __name__ == "__main__":
    unittest.main()



import struct
from jesth.tagbyte import TagByte
from jesth import errors


def deserialize(raw):
    deserializer = Deserializer(raw)
    return deserializer.run()


class Deserializer:
    def __init__(self, raw):
        self._raw = raw
        self._buffer = bytearray()
        self._data = None
        self._index = 0
        self._length = len(raw)
        self._setup()

    def run(self):
        containers = list()
        self._run(containers)
        return containers[0]

    def _setup(self):
        self._check_version()
        self._index = 1

    def _run(self, containers):
        length = len(self._raw)
        while self._index < length:
            tag = read_tag(self._raw, self._index)
            self._increment()
            size = 0
            if tag == TagByte.LIST_HEAD.value:
                new_container = list()
                if len(containers) != 0:
                    containers[-1].append(new_container)
                containers.append(new_container)
            elif tag == TagByte.LIST_TAIL.value:
                if len(containers) > 1:
                    del containers[-1]
            else:
                data, size = self._interpret_tag(tag)
                if containers:
                    containers[-1].append(data)
            self._increment(size)

    def _increment(self, n=1):
        self._index += n

    def _interpret_tag(self, tag):
        if tag == TagByte.INTEGER.value:
            data, size = read_integer(self._raw, self._index)
        elif tag == TagByte.LIST_HEAD.value:
            pass
        else:
            msg = "Unknown tag byte '{}'".format(tag)
            raise errors.Error(msg)
        return data, size

    def _check_version(self):
        version = extract_version(self._raw)
        if version == 1:
            pass
        else:
            msg = "This library can't deserialize this version of Jesth Binary Format"
            raise errors.Error(msg)


def read_tag(raw, index):
    spec = "<c"
    return struct.unpack_from(spec, raw, index)[0]


def read_integer(raw, index):
    spec = "<H"
    data = struct.unpack_from(spec, raw, index)[0]
    #data = int.from_bytes(data, "little")
    return data, struct.calcsize(spec)



def extract_version(raw):
    # extract version
    spec = "<B"
    r = struct.unpack_from(spec, raw, offset=0)
    return r[0]


"""Private miscellaneous functions. Only "split_key_value" is public"""
import os
import os.path
import pathlib
from tempfile import NamedTemporaryFile
from collections import namedtuple
from jesth import errors, const
try:
    import fcntl
except ImportError:
    fcntl = None


def write_to_file(data, dest):
    """
    Write to a file (TRY to do it atomically for GNU/Linux and MacOS)
    Check https://danluu.com/file-consistency/

    [parameters]
    - data: str, data to write to file
    - dest: a pathlib.Path instance, or a path string
    - binary_mode: boolean to tell if data should be written in dest with
      binary mode on (mode="wb")
    - encoding: defaults to "utf-8" and automatically set to None when
      binary mode is True
    """
    if isinstance(dest, pathlib.Path):
        dest = str(dest.resolve())
    if isinstance(dest, str):
        ensure_parent_dir(dest)
    else:
        msg = "Destination should be a path string or a pathlib.Path instance"
        raise errors.Error(msg)
    if os.path.isdir(dest):
        msg = "Destination shouldn't be a directory but a file path (which exists or not)"
        raise errors.Error(msg)
    WriteToFile.write(data, dest)


def split_key_value(line, sep="=", strip_whitespace=True):
    """
    Split a line (string) into key and value parts.
    The string must have a separator (defaults to "=").
    An example of a valid string is "name = John Doe ".
    The return will be: ("name": "John Doe")

    [parameters]
    - line: the string to split
    - sep: the separator, by default it is "="
    - strip_whitespace: boolean to tell whether you want whitespace at each
    side of the key and value to be stripped out or not.
    By default, the value is True.

    [return]
    Returns a ("key", "value") tuple if everything is Ok

    [exceptions]
    - errors.Error: if the separator, the key, or the value is missing,
    """
    cache = line.split(sep, maxsplit=1)
    if len(cache) != 2:
        msg = "Missing separator character '{}' in '{}'"
        msg = msg.format(sep, line)
        raise errors.Error(msg)
    key, value = cache
    if not key or key.isspace():
        msg = "Missing key in '{}'".format(line)
        raise errors.Error(msg)
    if not value or value.isspace():
        msg = "Missing value in '{}'".format(line)
        raise errors.Error(msg)
    if strip_whitespace:
        key, value = key.strip(), value.strip()
    return key, value


def correct_index(index, size, ignore_upper_bound=False):
    """
    Correct an index given to access an item in a list

    [parameter]
    - index: the integer representing the index of a list.
      This number may be positive or negative. A too big number is corrected down to
      the end index of the list.
      A negative number from -1 accesses the list from the end. A too small number
      is corrected up to 0.
      Example: give 5 the size of the list.
      Index -1 will be corrected to 4 and index 6 will be corrected to 4.
      Index -4 will be corrected to 0 same as index -10
    - size: an integer representing the size of the list

    [return]
    Return a corrected index integer
    """
    index = int(index)
    negative_size = -size
    end_index = size - 1
    if index >= size:
        if ignore_upper_bound:
            return index
        return end_index
    if index < negative_size:
        index = negative_size
    if index < 0:
        index = size - abs(index)
    return index


def ensure_parent_dir(path):
    """Make sure that parent dir exists (create it if isn't yet created)"""
    parent = os.path.dirname(path)
    try:
        os.makedirs(parent)
    except FileExistsError as e:
        pass

"""
def update_cached_refs(value, cached_refs):
    if value in cached_refs:
        raise errors.Error("Circular reference isn't allowed !")
    else:
        cached_refs.append(value)
"""

def count_indents(line):
    i = 0
    for char in line:
        if char == " ":
            i += 1
            continue
        break
    x = i / const.INDENT_WIDTH
    y = int(x)
    if x > y:
        raise errors.IndentError
    return y


def clean_leading_backslashes(line):
    """
    Remove one leading backslash only when the immediately-or-lately first non-backlash
    character on the line is an opening square bracket.
    """
    if not line or not line.startswith("\\"):
        return line
    if line.startswith(r"\["):
        line = line.replace(r"\[", "[", 1)
        return line
    char = None
    for i, char in enumerate(line):
        if char != "\\":
            break
    if char != "[":
        return line
    return line[1:]


def add_leading_backslashes(line):
    """
    Add one leading backslash only when the immediately-or-lately first non-backlash
    character on the line is an opening square bracket.
    """
    if not line:
        return line
    is_candidate = False
    for char in line:
        if char == "[":
            is_candidate = True
            break
        if char != "\\":
            break
    return "\\" + line if is_candidate else line


def get_headers(sections):
    "returns the set of headers from this list of sections"
    headers = set()
    for section in sections:
        headers.add(section.header)
    return headers


def decode_unicode(text):
    """Take in input some string that might have Unicode escape sequences.
    Output the same string with unicode escape sequences converted into
    the actual characters that they represent"""
    return text.encode("latin-1", "backslashreplace").decode("unicode-escape")


def encode_unicode(text, codec="latin-1"):
    """Convert a string into another where non-latin characters are
    replaced with Unicode escape sequences"""
    return text.encode(codec, "backslashreplace").decode(codec)


def tidy_up_float(s, width=3):
    """Tidy up a float number (str or float or decimal.Decimal)
    Example: 3.141234 -> 3.141_234
    """
    s = s if isinstance(s, str) else str(s)
    # make sure that s isn't "nan", "inf", "-inf", "infinity", ...
    for char in s:
        if char not in ("-01234.56789+eE"):
            return s
    info = parse_float(s)
    tidy_left_mantissa = tidy_up_int(info.left_mantissa)
    tidy_right_mantissa = ""
    if info.right_mantissa:
        cache = _tidy_up_right_mantissa(info.right_mantissa, width)
        tidy_right_mantissa = "." + cache
    tidy_exponent = "E" + info.exponent if info.exponent else ""
    result = tidy_left_mantissa + tidy_right_mantissa + tidy_exponent
    return result


def _tidy_up_right_mantissa(s, width):
    if not s:
        return ""
    cache = list()
    group = list()
    i = 0
    for char in s:
        if i == width:
            cache.append("".join(group))
            group = list()
            i = 0
        group.append(char)
        i += 1
    if group:
        cache.append("".join(group))
    return "_".join(cache)


def tidy_up_int(s, width=3):
    """Tidy up some int number (str or int)
    Example: 300141234 -> 300_141_234
    """
    s = s if isinstance(s, str) else str(s)
    tidy = list()
    is_negative = True if s[0] == "-" else False
    s = s.lstrip("+-")
    group = list()
    i = 0
    for char in reversed(s):
        if i == width:
            tidy.insert(0, "".join(group))
            group = list()
            i = 0
        group.insert(0, char)
        i += 1
    if group:
        tidy.insert(0, "".join(group))
    tidy = "_".join(tidy)
    result = "-" + tidy if is_negative else tidy
    return result


def tidy_up_hex(s, width=4):
    """Tidy up some hex int number (str or int)
    Example: 0xA3001F12A34 -> 0xA30_01F1_2A34
    """
    s = s if isinstance(s, str) else str(s)
    is_negative = True if s[0] == "-" else False
    s = s.lstrip("+-")
    s = s[2:]
    s = tidy_up_int(s, width=width)
    s = "0x" + s.upper()
    s = "-" + s if is_negative else s
    return s


def tidy_up_bin(s, width=4):
    """Tidy up some bin int number (str or int)
    Example: 0b10011101011 -> 0b100_1110_1011
    """
    s = s if isinstance(s, str) else str(s)
    is_negative = True if s[0] == "-" else False
    s = s.lstrip("+-")
    s = s[2:]
    s = tidy_up_int(s, width=width)
    s = "0b" + s
    s = "-" + s if is_negative else s
    return s


def tidy_up_oct(s, width=4):
    """Tidy up some oct int number (str or int)
    Example: 0o45631456202 -> 0o456_3145_6202
    """
    s = s if isinstance(s, str) else str(s)
    is_negative = True if s[0] == "-" else False
    s = s.lstrip("+-")
    s = s[2:]
    s = tidy_up_int(s, width=width)
    s = "0o" + s
    s = "-" + s if is_negative else s
    return s


FloatParts = namedtuple("FloatParts", ["left_mantissa", "right_mantissa", "exponent"])


def parse_float(s):
    """Parse a float number (string or decimal.Decimal or float), returns
    an instance of this namedtuple:
    FloatParts = namedtuple("FloatParts", ["left_mantissa", "right_mantissa", "exponent"])"""
    s = _prepare_float(s)
    # Split s at the exponent separator
    parts = s.split("E", 1)
    mantissa = parts[0]
    exponent = parts[1] if len(parts) == 2 else ""
    exponent = (exponent[1:] if exponent and exponent[0] == "+" 
                else exponent)
    # Split mantissa at the dot separator
    left_mantissa, right_mantissa = _parse_mantissa(mantissa)
    # Return the named tuple    
    return FloatParts(left_mantissa=left_mantissa, right_mantissa=right_mantissa, exponent=exponent)


def _prepare_float(s):
    if not isinstance(s, str):
        s = str(s)
    # Remove underscores, whitespaces, upper e
    cache = list()
    for c in s:
        if c in ("_", " "):
            continue
        if c == "e":
            c = "E"
        cache.append(c)
    return "".join(cache)


def _parse_mantissa(s):
    parts = s.split(".", 1)
    left_mantissa = parts[0]
    right_mantissa = parts[1] if len(parts) == 2 else ""
    return left_mantissa, right_mantissa


# --- write to file

class WriteToFile:
    @staticmethod
    def write(data, dest):
        # update encoding and mode
        binary_mode = True if isinstance(data, bytes) else False
        encoding = None if binary_mode else "utf-8"
        mode = "wb" if binary_mode else "w"
        # get parent dir and open file
        parent_dir = os.path.dirname(dest)
        # create temp file
        t = NamedTemporaryFile(prefix=".jesth_write_to_file_",
                               suffix=".temp", dir=parent_dir, delete=False)
        # write
        try:
            WriteToFile._write(data, t.name, mode, encoding)
            os.replace(t.name, dest)
            WriteToFile.sync_dir(parent_dir)
        except Exception as e:
            pass
        finally:
            try:
                os.unlink(t.name)
            except Exception as e:
                pass

    @staticmethod
    def _write(data, dest, mode, encoding):
        with open(dest, mode, encoding=encoding) as file:
            file.write(data)
            WriteToFile.sync_file(file)

    @staticmethod
    def sync_file(file):
        # file flush
        file.flush()
        # fsync the file
        fd = file.fileno()
        os.fsync(fd)
        WriteToFile.full_sync(fd)

    @staticmethod
    def sync_dir(path):
        # fsync parent dir also
        # https://man7.org/linux/man-pages/man2/fsync.2.html
        fd = os.open(path, os.O_RDONLY)
        try:
            os.fsync(fd)
            WriteToFile.full_sync(fd)
        except Exception as e:
            pass
        finally:
            os.close(fd)

    @staticmethod
    def full_sync(fd):
        # https://lists.apple.com/archives/darwin-dev/2005/Feb/msg00072.html
        if fcntl and hasattr(fcntl, "F_FULLFSYNC"):
            try:
                fcntl.fcntl(fd, fcntl.F_FULLFSYNC)
            except Exception as e:
                return False
            else:
                return True
        return False
